/*
 * Input output from TCP state listing and output PS code
 * $Id: timestamp.c,v 1.1 1997/03/04 09:08:48 puhuri Exp puhuri $
 * $Log: timestamp.c,v $
 * Revision 1.1  1997/03/04 09:08:48  puhuri
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#define FNAME_SIZE 1024
#define ERR_NOMEM 2

/* PS defines */
#define X_POINTS (72*200/25)
#define Y_POINTS (72*200/25)
#define LINELEN 100

#define i10E6 1000000		/* miljuuna */
#define DE_BYTES 1024*1024

typedef unsigned int  u32;

#define FLG_URG_C 1
#define FLG_URG (1<<FLG_URG_C)
#define FLG_ACK_C 2
#define FLG_ACK (1<<FLG_ACK_C)
#define FLG_PSH_C 3
#define FLG_PSH (1<<FLG_PSH_C)
#define FLG_RST_C 4
#define FLG_RST (1<<FLG_RST_C)
#define FLG_SYN_C 5
#define FLG_SYN (1<<FLG_SYN_C)
#define FLG_FIN_C 6
#define FLG_FIN (1<<FLG_FIN_C)

#define FLAGLEN 255

#define IN_FORMAT "%d.%d\t%u\t%u\t%u\t%u\t%[UAPRSF.]\n"

struct dataentry {
  long long usec;
  u32 seq, ack, win, mss;
  int flags;
};

struct datalist {
  struct datalist * next;
  struct dataentry *de;
  int count;
};

void chk_arch(void)
{
  if (sizeof(u32) != 4) { 
    fprintf (stderr, "u32 NOT 32 bit (%d bits)\n", 
	     sizeof(u32) * 8);
    exit (4);
  }
}

struct datalist * read_values (FILE *in, struct datalist *dlist)
{
  struct datalist *dl, *dll;
  dl = dlist;

  for (dll = dlist; dll; dll = dll->next)
    dl->count = 0;		/* reset counters */

  for (;;) {
    if (dl == NULL) {
      if ((dl = malloc (sizeof (struct datalist))) == NULL)
	exit(ERR_NOMEM);
      bzero (&dlist, sizeof (struct datalist));
    }
    if (dl->de == NULL)
      if ((dl->de = malloc (DE_BYTES)) == NULL)
	exit (ERR_NOMEM);

    for (dl->count = 0; dl->count < DE_BYTES / sizeof (struct dataentry);
	 dl->count++) {
      int sec, usec;
      struct dataentry *d;
      char flags[FLAGLEN];

      d = &(dl->de[dl->count]);

      if (fscanf (in, IN_FORMAT, &sec, &usec, &(d->seq),
		  &(d->ack), &(d->win),
		  &(d->mss), flags) < 7) {
	if (! feof(in))
	  fprintf (stderr, "Error parsing file\n");
	if (dl->count == 0)
	  fprintf (stderr, "GEE! just started new...");
	dl->count--;
	return dlist;
      }
      d->usec = usec;
      d->usec += i10E6 * (long long)sec;
      if (*flags == '.') {	/* shortcut */
	d->flags = 0;
      } else {
	d->flags =
	  ((index (flags, 'P')) ? FLG_PSH : 0) |
	  /* (index (flags, 'A')) ? FLG_URG : 0 |
	   * (index (flags, 'U')) ? FLG_URG : 0 |
	   */
	  ((index (flags, 'F')) ? FLG_FIN : 0) |
	  ((index (flags, 'S')) ? FLG_SYN : 0) |
	  ((index (flags, 'R')) ? FLG_RST : 0);
      }
    }
    dl = dl->next;
  }
}

static char fname[FNAME_SIZE+1];
static FILE *file_last_open;

FILE *open_ext (char *filename, char *ext)
{
  FILE *fout;

  strncpy (fname, filename, FNAME_SIZE);
  strncat (fname, "=", FNAME_SIZE); /* null terminated in any case? */
  strncat (fname, ext, FNAME_SIZE); /* null terminated in any case? */
  if ((fout = fopen (fname, "w")) == NULL) {
    fprintf (stderr,  "Open of \"%s\" failed: %s\n", fname,
	     strerror (sys_nerr));
    return NULL;
  }
  return file_last_open = fout;
}

void fclose_msg (FILE *fout)
{
  if (fclose (fout)) {
    fprintf (stderr,  "Close of \"%s\" failed: %s\n", 
	     (file_last_open == fout) ? fname : "*unknown*",
	     strerror (sys_nerr));
  }
}

int output_values (struct datalist *dlist, char *filename)
{
  int entries = 0;
  int counter = 0;
  struct datalist *d, *last = NULL;
  struct dataentry *de, *de_first, *de_last;
  u32 max_unack = 0;
  u32 min_unack = 100000;
  u32 max_win = 0;
  u32 min_win = 100000;
  int maxu = -1 , minu = -1 , maxw = -1, minw = -1;
  FILE *fout;
  long long usecs, base_time;
  u32 ack_scale;

  for (d = dlist; d; d = d->next) {
    entries += d->count + 1;
    last = d;
  }
  de_first = &(dlist->de[0]);
  de_last = &(last->de[last->count]);
  printf ("%s %d entries\n", filename, entries);
  printf ("Start: %d.%06d\n", (int)(de_first->usec / i10E6),
	  (int)(de_first->usec % i10E6));
  printf ("End:   %d.%06d\n", (int)(de_last->usec / i10E6),
	  (int)(de_last->usec % i10E6));
  base_time = de_first->usec;
  usecs = de_last->usec - base_time;

  for (d = dlist; d; d = d->next) {
    int i;
    de = d->de;

    for (i = 0; i <= d->count; i++){
      counter++;
      if (&(de[i]) == de_first || &(de[i]) == de_last)
	continue;		/* ignore 1st and last */
      if (de[i].flags & FLG_SYN)
	printf ("SYN @ %d\n", counter);
      if (de[i].seq + 1 < de[i].ack) {
	printf ("ACK %u > SEQ+1 %u\n", de[i].ack, de[i].seq + 1);
	continue;		/* ignore this */
      }
      if (max_unack < (de[i].seq + 1 - de[i].ack)) {
	maxu = counter;
	max_unack = (de[i].seq + 1 - de[i].ack);
      }
      if (min_unack > (de[i].seq + 1 - de[i].ack)){
	minu = counter;
	min_unack = (de[i].seq + 1 - de[i].ack);
      }
      if (max_win < (de[i].win)) {
	maxw = counter;
	max_win = (de[i].win);
      }
      if (min_win > (de[i].win)) {
	minw = counter;
	min_win = (de[i].win);
      }
    }
  }
  printf ("Window: %u -- %u (%d, %d)\n", min_win, max_win, minw, maxw);
  printf ("UNACK'd: %u -- %u (%d, %d)\n", min_unack, max_unack, minu, maxu);

  /* 
   * Number of unack bytes
   */
  if ((fout = open_ext (filename, "ack")) == NULL)
    return 0;

  fprintf (fout, 
	   "# %d points, %qd.%06qd sec Y=(%u .. %u)\n", 
	   entries, usecs / i10E6, usecs % i10E6,
	   min_unack, max_unack);
  fprintf (fout, "# Ack scale: %u .. %u\n", 
	   max_unack, min_unack);
  for (counter = 0, d = dlist; d; d = d->next) {
    int i;
    de = d->de;

    for (i = 0; i <= d->count; i++){
      long long now;
      u32 unack;
      counter++;
      now = de[i].usec - base_time;
      if (de[i].seq + 1 < de[i].ack)
	continue;		/* ignore this */
      unack = (de[i].seq + 1 - de[i].ack);
      if (unack > 65535)
	continue;
      fprintf (fout, "%qd.%06qd %u\n", now / i10E6, now % i10E6, unack);
    }
  }
  fclose_msg (fout);

  if ((fout = open_ext (filename, "tick")) == NULL)
    return 0;
  for (counter = 0, d = dlist; d; d = d->next) {
    int i;
    de = d->de;

    for (i = 0; i <= d->count; i++){
      long long now;
      counter++;
      now = de[i].usec - base_time;
      fprintf (fout, "%qd.%06qd %u\n", now / i10E6, now % i10E6,
	       counter % 2);
    }
  }

  fclose_msg (fout);

  if ((fout = open_ext (filename, "swin")) == NULL)
    return 0;

  ack_scale = max_win + max_unack - min_win - min_unack + 1;
  fprintf (fout, 
	   "# %d points, %qd.%06qd sec Y=(%u .. %u)\n", 
	   entries, usecs / i10E6, usecs % i10E6,
	   min_win - min_unack, max_unack + max_win);

  for (counter = 0, d = dlist; d; d = d->next) {
    int i;
    de = d->de;

    for (i = 0; i <= d->count; i++){
      long long now;
      u32 unack;
      counter++;
      now = de[i].usec - base_time;
      if (de[i].seq + 1 < de[i].ack)
	continue;		/* ignore this */
      unack = (de[i].seq + 1 - de[i].ack + de[i].win);
      if (unack > 65535 * 2)
	continue;
      fprintf (fout, "%qd.%06qd %u\n", now / i10E6, now % i10E6,
	       unack);
    }
  }
  fclose_msg(fout);

  do {
    long long dt;
    struct datalist *dl2;
    struct dataentry *de2;
    int dtc = 0;

    if ((fout = open_ext (filename, "bps")) == NULL)
      return 0;
    
    if (usecs > i10E6)
      dt = i10E6;
    else
      dt = usecs / 10;

    fprintf (fout, 
	     "# %d points, %qd.%06qd sec dt = %qd.%06qd\n", 
	     entries, usecs / i10E6, usecs % i10E6,
	     dt / i10E6, dt % i10E6
	     );
    
    for (counter = 0, dl2 = d = dlist, de2 = dl2->de; d; d = d->next) {
      int i;
      de = d->de;
      
      for (i = 0; i <= d->count; i++){
	long long now;
	counter++;

	now = de[i].usec - base_time;
	if (de2[dtc].usec < de[i].usec - dt) {
	  int dtc_t;
	  struct datalist *dl2_t;
	  struct dataentry *de2_t;

	  do {
	    dtc_t = dtc;
	    dl2_t = dl2;
	    de2_t = de2;
	    if (++dtc > dl2->count) {
	      dtc = 0; dl2 = dl2->next; de2 = dl2->de;
	    } 
	  } while (de2[dtc].usec < de[i].usec - dt);
	  dtc = dtc_t;
	  dl2 = dl2_t;
	  de2 = de2_t;
	}
	if (de[i].usec - de2[dtc].usec)
	  fprintf (fout, "%qd.%06qd %f\n", now / i10E6, now % i10E6,
		   (double)(de[i].seq - de2[dtc].seq) * i10E6 /
		   (double)(de[i].usec - de2[dtc].usec));
      }
    }
    fclose_msg(fout);
  } while (0);

  return 1;
}

int main (int argc, char **argv)
{
  int index;
  struct datalist dlist;
  FILE *in;

  chk_arch();
  bzero (&dlist, sizeof (struct datalist));
  for (index = 1; index < argc; index++) {
    if ((in = fopen (argv[index], "r")) == NULL) {
      fprintf (stderr,  "Open of \"%s\" failed: %s\n", argv[index],
	       strerror (sys_nerr));
      continue;
    }
    if (read_values(in, &dlist)){
      output_values (&dlist, argv[index]);
    }
    if (fclose (in)) {
      fprintf (stderr,  "Close of \"%s\" failed: %s\n", argv[index],
	       strerror (sys_nerr));
    }
  }
  return 0;
}
